import logging as log
from os import path
from pathlib import Path
import shutil

from .constants import OS, PWD, HOME_DIR, load_cfg
from .console_logging import print_header
from .console_logging import print_sub_header as sub_header
from .subproc import subproc


def rotate_github_ssh_keys() -> None:
    """
    update SSH pub keys for github.com
    """
    log.info("Rotating github.com ssh keys, just in case...")

    # create directory if it doesn't exist
    ssh_dir = path.join(HOME_DIR, '.ssh')
    if not path.exists(ssh_dir):
        Path(ssh_dir).mkdir(exist_ok=True)

    # create file if it doesn't exist
    known_hosts_file = path.join(ssh_dir, 'known_hosts')
    if not path.exists(ssh_dir):
        subproc([f"touch {known_hosts_file}"])

    # deletes all keys starting with github.com from ~/.ssh/known_hosts
    subproc(["ssh-keygen -R github.com"])

    # gets the new public keys from github.com
    github_keys = subproc(["ssh-keyscan github.com"])

    # the new github.com keys are not automatically added :( so we do it here
    with open(known_hosts_file, 'a') as known_hosts:
        for line in github_keys.split('/n'):
            known_hosts.write(line)


def run_preinstall_cmds(cmd_list: list, 
                        pkg_groups: list, 
                        no_upgrade: bool) -> None:
    """
    takes a list of package manager pre-install commands and runs them
    if second list of package groups contains gaming, runs additional commands
    returns True
    """
    run_gaming_setup = False
    if 'gaming' in pkg_groups:
        run_gaming_setup = True

    pre_cmds = ['setup', 'update', 'upgrade']

    if no_upgrade:
        pre_cmds.pop(2)

    for pre_cmd in pre_cmds:
        if pre_cmd in cmd_list:
            if pre_cmd == 'update' and 'apt' in cmd_list[pre_cmd]:
                if run_gaming_setup:
                    log.debug("Run gaming commands to update /etc/apt/sources")
                    cmds = ["sudo dpkg --add-architecture i386",
                            f"sudo {PWD}/scripts/update_apt_sources.sh"]
                    subproc(cmds, spinner=False)
            if pre_cmd == 'upgrade' and 'brew' in cmd_list[pre_cmd]:
                if OS[0] == 'Darwin':
                    log.debug("we'll run upgrade --cask for macOS")
                    subproc(["brew upgrade --cask"])

            SPINNER = True
            if 'sudo' in cmd_list[pre_cmd]:
                SPINNER = False

            subproc([cmd_list[pre_cmd]], spinner=SPINNER)
            sub_header(f"[b]{pre_cmd.title()}[/b] completed.")


def run_pkg_mngrs(pkg_mngrs: list, pkg_groups=[], no_upgrade=bool) -> None:
    """
    Installs brew and pip3.11 packages. Also apt, snap, and flatpak on Linux.
    Takes optional variables:
      - pkg_groups: list of optional package groups
      - pkg_mngrs: list of package managers to run
    Returns True
    """
    log.debug(f"passed in pkg_mngrs: {pkg_mngrs}\npkg_groups: {pkg_groups}")

    rotate_github_ssh_keys()

    pkg_mngrs_list_of_dicts = load_cfg('packages.yml')
    # we iterate through pkg_mngrs which should already be sorted
    for pkg_mngr in pkg_mngrs:
        pkg_mngr_dict = pkg_mngrs_list_of_dicts[pkg_mngr]

        available_pkg_groups = pkg_mngr_dict['packages']
        log.debug(f"pkg groups for {pkg_mngr} are {available_pkg_groups}")
        
        # brew has a special flow because it works on both linux and mac
        if pkg_mngr == 'brew':
            if 'Darwin' in OS:
                if 'default' in pkg_groups:
                    if type(pkg_groups) is tuple:
                        pkg_groups = list(pkg_groups)
                    pkg_groups.append("macOS")

        # make sure that the package manager has any groups that were passed in
        if any(check in pkg_groups for check in available_pkg_groups):

            pkg_emoji = pkg_mngr_dict['emoji']
            msg = f'{pkg_emoji} [grn][b]{pkg_mngr}[/b][/grn] app management'
            print_header(msg)

            # commands for listing, installing, updating, upgrading, & cleanup
            pkg_cmds = pkg_mngr_dict['commands']

            # make sure we use any required env vars during installs
            install_env_vars = pkg_mngr_dict.get('env_vars', None)

            if pkg_mngr == 'snap' and not shutil.which('snap'):
                log.warn("snap is either not installed, or you need to log out"
                         "and back in (or reboot) for it to be available. "
                         "https://snapcraft.io/docs/installing-snap-on-debian")
                # continues onto the next package manager
                continue

            if pkg_mngr == 'brew':
                if 'Darwin' in OS:
                    # this installs macOS brew taps
                    install_brew_taps(pkg_mngr_dict['taps']['macOS'])

            # run package manager specific setup if needed: update/upgrade
            run_preinstall_cmds(pkg_cmds, pkg_groups, no_upgrade)

            # run the list command for the given package manager
            list_cmd = pkg_cmds['list']
            # TODO: figure out a way to make this less hacky
            if pkg_mngr == 'apt':
                list_cmd = path.join(PWD, list_cmd)
            list_pkgs = subproc([list_cmd], quiet=True)

            if list_pkgs:
                # create list of installed packages to iterate on
                installed_pkgs = list_pkgs.split()
            else:
                installed_pkgs = []

            # iterate through package groups for a given package manager
            for pkg_group in pkg_groups:
                # if package group is in the packages.yaml file
                if pkg_group in available_pkg_groups:
                    # used to be for zathura, a document viewer, might delete
                    # if pkg_group == "macOS":
                    #     # zathura needs some help on macOS
                    #     check_zathura()

                    install_pkg_group(pkg_cmds['install'],
                                      available_pkg_groups[pkg_group],
                                      installed_pkgs,
                                      install_env_vars,
                                      no_upgrade)
                    sub_header(f'{pkg_group.title()} packages installed.')

            # run final cleanup commands, if any
            if 'cleanup' in pkg_cmds:
                subproc([pkg_cmds['cleanup']])
                sub_header("[b]Cleanup[/b] step Completed.")


def install_pkg_group(install_cmd: str, 
                      pkgs_to_install: list,
                      installed_pkgs: list, 
                      install_env_vars: dict,
                      no_upgrade: bool
                      ) -> None:
    """
    Installs packages if they are not already installed.
    provided install command string.
    Returns True.
    """
    log.debug(f"Currently installed packages: {installed_pkgs}")
    log.info(f"Packages to install are: {pkgs_to_install}")

    for pkg in pkgs_to_install:
        if not installed_pkgs:
            log.info(f"{pkg} isn't installed. Installing now...")
        else:
            # this variable defaults to pkg unless it has special strings in it
            pkg_short_name = pkg

            # for things like steam:i386 for apt
            if ":" in pkg:
                pkg_short_name = pkg.split(':')[0]

            # this covers things like "--cask iterm2" for brew
            if "--cask" in pkg:
                pkg_short_name = pkg.split(' ')[1]

            log.debug(f"Checking if {pkg_short_name} is installed...")
            if pkg_short_name in installed_pkgs:
                if no_upgrade or 'upgrade' not in install_cmd:
                    log.info(f"{pkg} already installed. Moving on.")
                    # continues to the next pkg in the pkgs_to_install list
                    continue
                # if the install command has upgrade in it, we always run it
                else:
                    log.info(f"Upgrading {pkg} now...")

        # Actual installation
        subproc([install_cmd + pkg], quiet=True, env=install_env_vars)


def install_brew_taps(taps: list) -> None:
    """
    Checks current brew taps, and then runs brew tap {tap} on any taps that are
    in a list of git repos from packages.yaml, and aren't already tapped
    """
    log.info("Checking current taps.")
    existing_taps = subproc(["brew tap"])
    try:
        current_taps = existing_taps.split('\n')
    except Exception as e:
        log.debug(e)
        current_taps = []
    
    log.debug(f"taps list is: {taps}")
    log.debug(f"Current taps are: {current_taps}")

    # for each tap, complete cmd by prepending `brew tap`
    for index, tap in enumerate(taps):
        log.debug(f"index: {index} tap: {tap}")
        # only brew tap if they don't already exist
        if tap not in current_taps:
            subproc(["brew tap " + tap])
        else:
            log.info(f"{tap} already installed. Moving on.")


# not currently using zathura on macOS, so removing as it's untested
# def check_zathura() -> None:
#     """
#     make sure zathura is installed on macos
#     installs via brew if it's not installed
#     always returns True if everything was successful
#     """
#     if not shutil.which("zathura"):
#         zathura_pdf = "$(brew --prefix zathura-pdf-mupdf)"
#         cmds = ["mkdir -p $(brew --prefix zathura)/lib/zathura",
#                 f"ln -s {zathura_pdf}/libpdf-mupdf.dylib" +
#                 f"{zathura_pdf}/lib/zathura/libpdf-mupdf.dylib"]
#         subproc(cmds, quiet=True)



"""
    NAME:           onboardme
    DESCRIPTION:    Program to take care of a bunch of onboarding tasks for new
                    machines running macOS and/or Debian, including:
                      dot_files, managing packages, ide_setup, group management
    AUTHOR:         Jesse Hitch
    LICENSE:        GNU AFFERO GENERAL PUBLIC LICENSE
"""

from click import option, command, Choice
from importlib import import_module
import logging
from rich.logging import RichHandler
from .help_text import RichCommand, options_help
from .constants import VERSION, OS, STEPS, PKG_MNGRS, USR_CONFIG_FILE
from .env_config import check_os_support, process_configs
from .console_logging import print_manual_steps
from .dot_files import setup_dot_files
from .pkg_management import run_pkg_mngrs
from .sudo_setup import setup_sudo
from .firewall import configure_firewall


HELP = options_help()


def setup_logger(level="", log_file=""):
    """
    Sets up rich logger and stores the values for it in a db for future import
    in other files. Returns logging.getLogger("rich")
    """
    # determine logging level
    if not level:
        if USR_CONFIG_FILE and 'log' in USR_CONFIG_FILE:
            level = USR_CONFIG_FILE['log']['level']
        else:
            level = 'warn'

    log_level = getattr(logging, level.upper(), None)

    # these are params to be passed into logging.basicConfig
    opts = {'level': log_level, 'format': "%(message)s", 'datefmt': "[%X]"}

    # we only log to a file if one was passed into config.yml or the cli
    if not log_file:
        if USR_CONFIG_FILE:
            log_file = USR_CONFIG_FILE['log'].get('file', None)

    # rich typically handles much of this but we don't use rich with files
    if log_file:
        opts['filename'] = log_file
        opts['format'] = "%(asctime)s %(levelname)s %(funcName)s: %(message)s"
    else:
        rich_handler_opts = {'rich_tracebacks': True}
        # 10 is the DEBUG logging level int value
        if log_level == 10:
            # log the name of the function if we're in debug mode :)
            opts['format'] = "[bold]%(funcName)s()[/bold]: %(message)s"
            rich_handler_opts['markup'] = True

        opts['handlers'] = [RichHandler(**rich_handler_opts)]

    # this uses the opts dictionary as parameters to logging.basicConfig()
    logging.basicConfig(**opts)

    if log_file:
        return None
    else:
        return logging.getLogger("rich")


# @option('--quiet', '-q', is_flag=True, help=HELP['quiet'])
# Click is so ugly, and I'm sorry we're using it for cli parameters here, but
# this allows us to use rich.click for pretty prettying the help interface
# each of these is an option in the cli and variable we use later on
@command(cls=RichCommand)
@option('--log_level', '-l', metavar='LOGLEVEL', help=HELP['log_level'],
        type=Choice(['debug', 'info', 'warn', 'error']),
        default=USR_CONFIG_FILE['log']['level'])
@option('--log_file', '-o', metavar='LOGFILE', help=HELP['log_file'],
        default=USR_CONFIG_FILE['log']['file'])
@option('--steps', '-s', metavar='STEP', multiple=True, type=Choice(STEPS),
        help=HELP['steps'], default=USR_CONFIG_FILE['steps'][OS[0]])
@option('--git_url', '-u', metavar='URL', help=HELP['git_url'],
        default=USR_CONFIG_FILE['dot_files']['git_url'])
@option('--git_branch', '-b', metavar='BRANCH', help=HELP['git_branch'],
        default=USR_CONFIG_FILE['dot_files']['git_branch'])
@option('--overwrite', '-O', is_flag=True, help=HELP['overwrite'],
        default=USR_CONFIG_FILE['dot_files']['overwrite'])
@option('--pkg_managers', '-p', metavar='PKG_MANAGER', multiple=True,
        type=Choice(PKG_MNGRS), help=HELP['pkg_managers'],
        default=USR_CONFIG_FILE['package']['managers'][OS[0]])
@option('--pkg_groups', '-g', metavar='PKG_GROUP', multiple=True,
        type=Choice(['default', 'gaming', 'gui', 'devops']),
        help=HELP['pkg_groups'], default=USR_CONFIG_FILE['package']['groups'])
@option('--firewall', '-f', is_flag=True, help=HELP['firewall'],
        default=USR_CONFIG_FILE.get('firewall', False))
@option('--remote_host', '-r', metavar="IP_ADDR", multiple=True,
        help=HELP['remote_host'],
        default=USR_CONFIG_FILE.get('remote_hosts', None))
@option('--no_upgrade', '-n', help=HELP['no_upgrade'], default=False, 
        is_flag=True)
@option('--version', is_flag=True, help=HELP['version'], default=False)
def main(log_level, log_file,
         steps, 
         git_url, git_branch, overwrite,
         pkg_managers, pkg_groups,
         firewall, remote_host,
         no_upgrade,
         version) -> bool:
    """
    If run with no options on Linux, it will install brew, pip3.11, apt,
    flatpak, and snap packages. On mac, it only installs brew/pip3.11 packages.
    config loading tries to load: cli options and then defaults back to:
    $XDG_CONFIG_HOME/onboardme/config.yml
    """

    # only return the version if --version was passed in
    if version:
        print(f'\n🎉 v{VERSION}\n')
        return True

    # before we do anything, we need to make sure this OS is supported
    check_os_support()

    # setup logging immediately
    log = setup_logger(log_level, log_file)

    # makes sure we only overwrite config file prefs if cli opts are passed in
    usr_pref = process_configs(overwrite, git_url, git_branch, pkg_managers,
                               pkg_groups, firewall, remote_host, steps,
                               log_file, log_level)

    if log:
        log.debug(f"User passed in the following preferences:\n{usr_pref}\n")
    else:
        logging.debug(f"User passed in the following preferences:\n{usr_pref}")

    # actual heavy lifting of onboardme happens in these
    for step in usr_pref['steps'][OS[0]]:

        if step == 'dot_files':
            # this creates a live git repo out of your home directory
            df_prefs = usr_pref['dot_files']
            setup_dot_files(OS, df_prefs['overwrite'],
                            df_prefs['git_url'], df_prefs['git_branch'])

        elif step == 'packages':
            pkg_mngrs = usr_pref['package']['managers'][OS[0]]
            pkg_groups = usr_pref['package']['groups']
            run_pkg_mngrs(pkg_mngrs, pkg_groups, no_upgrade)

        elif step in ['neovim_setup', 'font_setup']:
            # import step's function from ide_setup.py in same directory
            import_module('onboardme.ide_setup', package=f'.{step}')
            func = getattr(ide_setup, step)
            func()

        elif step == 'sudo_setup':
            # if we're not running as root, kick off another process
            setup_sudo()

    if 'firewall_setup' in steps:
        configure_firewall(remote_host)

    print_manual_steps(OS)
    return True


if __name__ == '__main__':
    main()
